package ru.dbotthepony.mc.otm.android

import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import net.minecraft.ChatFormatting
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.network.chat.Component
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.ItemStack
import net.minecraftforge.common.util.INBTSerializable
import ru.dbotthepony.mc.otm.capability.MatteryPlayerCapability
import ru.dbotthepony.mc.otm.capability.itemsStream
import ru.dbotthepony.mc.otm.client.render.SkinElement
import ru.dbotthepony.mc.otm.container.iterator
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.addAll
import ru.dbotthepony.mc.otm.core.getCompoundList
import ru.dbotthepony.mc.otm.core.nonEmpty
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.network.FieldSynchronizer
import ru.dbotthepony.mc.otm.registry.MRegistry
import java.io.DataInputStream
import java.io.InputStream

class AndroidResearch(val type: AndroidResearchType, val capability: MatteryPlayerCapability) : INBTSerializable<CompoundTag> {
	val ply: Player get() = capability.ply

	val synchronizer = FieldSynchronizer()

	var isResearched by synchronizer.bool()
		private set

	/**
	 * Called when it is required to network everything again
	 */
	fun invalidateNetwork() {
		synchronizer.invalidate()
	}

	fun unResearch() {
		if (!isResearched) {
			return
		}

		onUnResearch()
		isResearched = false
	}

	private data class RememberResearchLevel(val level: Int?)

	private val oldResearchLevel = Object2ObjectArrayMap<AndroidFeatureType<*>, RememberResearchLevel>()

	fun onUnResearch() {
		for (feature in type.resolvedFeatures) {
			val level = oldResearchLevel[feature.feature]
			val get = capability.getFeature(feature.feature)

			if (level != null && get != null) {
				if (get.level == feature.level) {
					if (level.level == null) {
						capability.removeFeature(feature.feature)
					} else {
						get.level = level.level

						for (transformer in type.features.first { it.id == feature.feature.registryName }.transformersDown) {
							transformer.accept(this to get)
						}
					}
				}
			}
		}

		oldResearchLevel.clear()
	}

	fun onResearched() {
		oldResearchLevel.clear()

		try {
			for (feature in type.resolvedFeatures) {
				var get = capability.getFeature(feature.feature)

				if (get == null) {
					get = capability.addFeature(feature.feature)
					get.level = feature.level
					oldResearchLevel[feature.feature] = RememberResearchLevel(null)
				} else {
					if (get.level < feature.level) {
						oldResearchLevel[feature.feature] = RememberResearchLevel(feature.level)
						get.level = feature.level
					}
				}

				for (transformer in type.features.first { it.id == feature.feature.registryName }.transformersUp) {
					transformer.accept(this to get)
				}
			}
		} catch(err: Throwable) {
			oldResearchLevel.clear()
			throw err
		}
	}

	/**
	 * Consumes required resources for researching this technology.
	 *
	 * Returns true if all required resources were consumed, false otherwise.
	 *
	 * This function MUST be atomic, hence it SHOULD NOT consume ANY resources if it returns false.
	 */
	fun consumeResearchCost(simulate: Boolean): Boolean {
		if (capability.ply.abilities.instabuild) {
			// creative
			return true
		}

		if (!simulate && !consumeResearchCost(true)) {
			return false
		}

		if (type.experienceLevels > 0) {
			if (capability.ply.experienceLevel < type.experienceLevels) {
				return false
			}

			if (!simulate) {
				capability.ply.giveExperienceLevels(-type.experienceLevels)
			}
		}

		for (item in type.items) {
			var required = item.count
			val iterator = capability.ply.inventory.iterator().nonEmpty()

			for (invItem in iterator) {
				if (ItemStack.isSameItemSameTags(invItem, item)) {
					val toExtract = required.coerceAtMost(invItem.count)
					required -= toExtract

					if (!simulate) {
						if (toExtract == invItem.count) {
							iterator.remove()
						} else {
							invItem.count -= toExtract
						}
					}

					if (required <= 0) {
						break
					}
				}
			}

			if (required > 0) {
				return false
			}
		}

		return true
	}

	/**
	 * Grants all (or some) resources back consumed by [consumeResearchCost].
	 *
	 * Returns true whenever player accepted all resources refunded, false otherwise.
	 */
	fun refund(simulate: Boolean): Boolean {
		if (simulate) {
			return true
		}

		if (type.experienceLevels > 0) {
			capability.ply.giveExperienceLevels(type.experienceLevels)
		}

		for (item in type.items) {
			capability.ply.inventory.add(item)

			if (!item.isEmpty) {
				capability.ply.spawnAtLocation(item)
			}
		}

		return true
	}

	fun collectNetworkPayload(): FastByteArrayOutputStream? {
		return synchronizer.collectNetworkPayload()
	}

	fun applyNetworkPayload(stream: InputStream) {
		synchronizer.applyNetworkPayload(DataInputStream(stream))
	}

	val canResearch: Boolean get() {
		if (!consumeResearchCost(simulate = true)) {
			return false
		}

		if (!isPrerequisitesResearched || (!ply.isCreative && isAnyBlockerResearched)) {
			return false
		}

		return true
	}

	/**
	 * Determines if this research is already blocked directly by another research.
	 */
	val isAnyBlockerResearched: Boolean get() {
		for (research in type.flatBlockedBy) {
			if (capability.getResearch(research).isResearched) {
				return true
			}
		}

		return false
	}

	/**
	 * Determines if this research is already blocked by another research, directly or indirectly.
	 *
	 * This is slower than [isAnyBlockerResearched] and makes sense only to check this for visualization purposes,
	 * since indirect blocking research will just turn [isPrerequisitesResearched] false (under normal conditions).
	 */
	val isAnyBlockerResearchedIndirect: Boolean get() {
		for (research in type.allBlockedBy) {
			if (capability.getResearch(research).isResearched) {
				return true
			}
		}

		return false
	}

	/**
	 * Determines if this research' direct prerequisites are researched.
	 */
	val isPrerequisitesResearched: Boolean get() {
		for (research in type.flatPrerequisites) {
			if (!capability.getResearch(research).isResearched) {
				return false
			}
		}

		return true
	}

	/**
	 * Determines if this research' prerequisites are researched, direct or indirect.
	 *
	 * Doesn't make much sense to check for this, since all prerequisites *should* have their prerequisites
	 * researched and so on.
	 *
	 * Strictly speaking, if someone tamper with research tree and mark all direct prerequisites unlocked
	 * without unlocking their prerequisites, [isPrerequisitesResearched] will return true, but *this* will return *false*.
	 */
	val isPrerequisitesResearchedIndirect: Boolean get() {
		for (research in type.allPrerequisites) {
			if (!capability.getResearch(research).isResearched) {
				return false
			}
		}

		return true
	}

	fun research(force: Boolean = false): Boolean {
		if (isResearched) {
			return false
		}

		if (force || canResearch && consumeResearchCost(false)) {
			onResearched()
			isResearched = true
			return true
		}

		return false
	}

	/**
	 * List of all tooltip lines for this research
	 */
	val tooltipLines: List<Component> get() {
		val lines = ArrayList<Component>()
		lines.add(type.displayName)
		lines.addAll(type.description.iterator())

		return lines
	}

	/**
	 * List of all tooltip lines for this research, for local player at research screen (clientside only)
	 */
	val screenTooltipLines: List<Component> get() {
		val builder = tooltipLines as ArrayList<Component>

		if (type.experienceLevels != 0) {
			builder.add(
				TranslatableComponent("otm.android_station.research.xp_cost", type.experienceLevels).withStyle(
					if (capability.ply.experienceLevel >= type.experienceLevels)
						ChatFormatting.DARK_GREEN
					else
						ChatFormatting.DARK_RED
				))
		}

		for (value in this.type.flatPrerequisites) {
			val instance = capability.getResearch(value)

			builder.add(
				TranslatableComponent("android_research.status.requires", instance.screenTooltipHeader).withStyle(
					if (instance.isResearched)
						ChatFormatting.DARK_GREEN
					else
						ChatFormatting.DARK_RED
				))
		}

		for (value in this.type.flatBlockedBy) {
			builder.add(TranslatableComponent("android_research.status.blocked_by", capability.getResearch(value).screenTooltipHeader).withStyle(ChatFormatting.DARK_RED))
		}

		for (value in this.type.flatBlocking) {
			builder.add(TranslatableComponent("android_research.status.blocks", capability.getResearch(value).screenTooltipHeader).withStyle(ChatFormatting.DARK_RED))
		}

		return builder
	}

	/**
	 * This should return whatever can be used by other code to display this research's name,
	 * for local player at research screen (clientside only)
	 */
	val screenTooltipHeader: Component get() = type.displayName

	override fun serializeNBT(): CompoundTag {
		return CompoundTag().also {
			it["researched"] = isResearched

			it["oldResearchLevel"] = ListTag().also {
				for ((k, v) in oldResearchLevel) {
					it.add(CompoundTag().also {
						it["key"] = k.registryName!!.toString()
						it["value"] = CompoundTag().also {
							it["isPresent"] = v.level != null

							if (v.level != null) {
								it["value"] = v.level
							}
						}
					})
				}
			}
		}
	}

	override fun deserializeNBT(nbt: CompoundTag) {
		isResearched = nbt.getBoolean("researched")

		oldResearchLevel.clear()

		for (tag in nbt.getCompoundList("oldResearchLevel")) {
			val key = tag.getString("key")
			val type = MRegistry.ANDROID_FEATURES.getValue(ResourceLocation(key)) ?: continue
			val value = tag.getCompound("value")

			val isPresent = value.getBoolean("isPresent")
			val int = value.getInt("value")

			oldResearchLevel[type] = RememberResearchLevel(if (isPresent) int else null)
		}
	}
}
